// Code generated by go generate; DO NOT EDIT.
package network

import (
	"net/url"

	"github.com/containers/podman/v3/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *ConnectOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *ConnectOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithAliases set field Aliases to given value
func (o *ConnectOptions) WithAliases(value []string) *ConnectOptions {
	o.Aliases = &value
	return o
}

// GetAliases returns value of field Aliases
func (o *ConnectOptions) GetAliases() []string {
	if o.Aliases == nil {
		var z []string
		return z
	}
	return *o.Aliases
}
